=begin
Copyright 2014, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Author:       Andreas Eisenbarth

Usage:        Menu 'Plugins' → '[Test] PNG Loader Tool'
              Click to switch from one test image to another one.
              Double-click to load your own png image.

Description:  This is a test for rendering images on the SketchUp viewport.
              The idea is to have only one drawing operation per color, and pass 
              each time the coordinates of all points of that same color.
              This is different from the internal structure of image files 
              (which store a sequence of colors whereas the coordinates are obvious),
              so we need to sort the pixels by color and store there cordinates.

Required:     SketchUp 2014 with standard lib for loading ChunkyPNG.

Version:      1.0.2

Date:         28.06.2014


TODO: embed images with alpha transparency. Since ChunkyPNg cannot load such files, we need to save all channels to individual black&white pngs and merge them by script.

=end
module AE


module PNGLoader


DIR = File.dirname(__FILE__) unless defined?(self::DIR)
begin
  require File.join(DIR, "chunky_png.rb")
rescue LoadError
rescue Exception
end


class PNGLoaderTool


def initialize
  # Image data: The internal format is a hash where to each color an array of coordinates is assigned.
  #   { Sketchup::Color => [[Fixnum,Fixnum],...] }
  @colors = {}
  # Statistical data
  @basename = ""
  @file_loaded_time = 0
  @pixel_count = 0
  @frames = 0
  @frames_dt = 0
  # Image switching
  @images = ["dmp/sketchup100.dmp", "dmp/sketchup200.dmp", "dmp/trimble.dmp", "dmp/streetcar.dmp"]
  @state = 0
  path = File.join(DIR, @images[@state])
  load_rb(path)
  # 
  Sketchup.set_status_text("Click to change the image, double-click to load a png file.")
  Sketchup.active_model.active_view.invalidate
end


def onLButtonDown(flags, x, y, view)
  # Cycle through the states
  @state = (@state + 1) % @images.length
  # And load the corresponding image
  path = File.join(DIR, @images[@state])
  load_rb(path)
  view.invalidate
end


def onLButtonDoubleClick(flags, x, y, view)
  # Try to load a png file
  path = UI.openpanel("Select a png file")
  load_png(path) rescue UI.messagebox("Loading failed. The png library seems to support only some png files, try creating the png file with 'index colors', 'web save colors', 'Floyd-Steinberg rasterization'.")
  view.invalidate
end


#private


def load_png(path=nil)
  return if path.nil? || !File.exists?(path) || !defined?(ChunkyPNG)
  t0 = Time.now
  # Read the image
  image = ChunkyPNG::Image.from_file(path)
  @file_name = File.basename(path)
  @pixel_count = image.pixels.length
  @width = image.width
  @height = image.height
  # Sort by color and store the coordinates.
  @colors = {}
  image.pixels.each_with_index{ |p, i|
    # Look up if the color has already been added to @colors. Could be optimized.
    r = ChunkyPNG::Color.r(p)
    g = ChunkyPNG::Color.g(p)
    b = ChunkyPNG::Color.b(p)
    a = ChunkyPNG::Color.a(p)
    color = @colors.keys.find{ |c| c.red == r && c.green == g && c.blue == b && c.alpha == a }
    # Otherwise create a new color.
    color = Sketchup::Color.new(r, g, b, a) if color.nil?
    @colors[color] ||= []
    # Get the coordinates
    y, x = i.divmod(@width)
    # Store the coordinates for this color.
    @colors[color] << [x+1, y+1]
  }
  @file_loaded_time = (Time.now - t0).to_f
  @frames = 0
  @frames_dt = 0
end


def load_rb(path=UI.openpanel)
  return if path.nil? || !File.exists?(path)
  @file_name = File.basename(path)
  t0 = Time.now
  @colors = {}
  File.open(path, "rb"){ |f|
    Marshal.load(f.read).each{ |int, array|
      @colors[Sketchup::Color.new(int)] = array
    }
  }
  @file_loaded_time = (Time.now - t0).to_f
  @pixel_count = @colors.values.inject(0){ |sum, array| sum + array.length }
  @width = @colors.values.map{|a| a.map{|p| p[0] } }.flatten.max
  @height = @pixel_count / @width
  @frames = 0
  @frames_dt = 0
end


def save_rb(path=UI.savepanel)
  return if path.nil?
  File.open(path, "wb"){ |f|
    # Sketchup::Color can not be marshalled, so we take the [r,g,b,a] array.
    hash = {}
    @colors.each{ |color, array|
      hash[color.to_a] = array
    }
    f.write Marshal.dump(hash)
  }
end


def draw(view)
  t0 = Time.now
  # Draw for each color all points.
  @colors.each{ |color, points|
    view.drawing_color = color#Sketchup::Color.new(*color)
    view.draw2d(GL_POINTS, points)
  }
  @frames += 1
  @frames_dt += (Time.now - t0).to_f
  # Draw benchmarking data.
  view.draw_text([0,@height], "#{@file_name} (#{@width}×#{@height}px, loaded in #{@file_loaded_time} seconds)\n" + 
  "Drawing #{@pixel_count} pixels of #{@colors.length} colors in #{sprintf("%2.4f", @frames_dt/@frames.to_f)} seconds")
end


end # class PNGLoaderTool


unless file_loaded?(__FILE__)
  command = UI::Command.new("[Test] PNG Loader Tool"){
    Sketchup.active_model.select_tool(AE::PNGLoader::PNGLoaderTool.new)
  }
  UI.menu("Plugins").add_item(command)
  file_loaded(__FILE__)
end


end # module PNGLoader


end # module AE